/*
 *	SuperTrak EX8350/8300/16350/16300 Storage Controller device driver for Linux
 *
 * Copyright (C) 2005 Promise Technology Inc.
 * 
 *	   This program is free software; you can redistribute it and/or
 *	   modify it under the terms of the GNU General Public License
 *	   as published by the Free Software Foundation; either version
 *	   2 of the License, or (at your option) any later version.
 *
 * Written By	: 
 *	Ed Lin <edl@promisechina.com>
 *
 * Version	: 2.9.0.10
 *
 */
#include <linux/init.h>
#include <stdarg.h>
#include <asm/io.h>
#include <asm/string.h>
#include <asm/irq.h>
#include <asm/byteorder.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/irq.h>
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <linux/blk.h>
#else
#include <linux/blkdev.h>
#include <linux/interrupt.h>
#endif
#include <linux/types.h>
#include <linux/hdreg.h>
#include <linux/notifier.h>        
#include <linux/reboot.h>

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/smp.h>

#include <asm/uaccess.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <scsi.h>
#include <sd.h>
#include <hosts.h>
#else
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_host.h>
#endif
#ifdef _64BPLATFORM
#include <linux/ioctl32.h>
#endif

#define SHASTA_NAME "shasta"
#define ST_DRIVER_DATE "(Dec 19, 2005)"
#ifdef _NEC
#define OEM 	70
#define ST_DRIVER_VERSION "2.9.0070.10"
#else
#define OEM 	0
#define ST_DRIVER_VERSION "2.9.0.10"
#endif
#define VER_MAJOR 		2	
#define VER_MINOR 		9
#define BUILD_VER 		10

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,5)
#ifndef __iomem
#define __iomem
#endif

#define msleep(x) udelay(x*1000)

typedef u16 __le16;
typedef u32 __le32;

#endif

enum{
	/* MU status code */
	MU_STATE_STARTING			= 1,
	MU_STATE_FMU_READY_FOR_HANDSHAKE	= 2,
	MU_STATE_SEND_HANDSHAKE_FRAME		= 3,
	MU_STATE_STARTED			= 4,

	MU_MAX_DELAY_TIME			= 10000,
	MU_HANDSHAKE_SIGNATURE			= 0x55aaaa55,
	HMU_PARTNER_TYPE			= 2,

	/* MU registers */
	IMR0	= 0x10,	/* MU_INBOUND_MESSAGE_REG0 */
	IMR1	= 0x14,	/* MU_INBOUND_MESSAGE_REG1 */
	OMR0	= 0x18,	/* MU_OUTBOUND_MESSAGE_REG0 */
	OMR1	= 0x1c,	/* MU_OUTBOUND_MESSAGE_REG1 */
	IDBL	= 0x20,	/* MU_INBOUND_DOORBELL */
	IIS	= 0x24,	/* MU_INBOUND_INTERRUPT_STATUS */
	IIM	= 0x28,	/* MU_INBOUND_INTERRUPT_MASK */
	ODBL	= 0x2c,	/* MU_OUTBOUND_DOORBELL */
	OIS	= 0x30,	/* MU_OUTBOUND_INTERRUPT_STATUS */
	OIM	= 0x3c,	/* MU_OUTBOUND_INTERRUPT_MASK */

	MU_INBOUND_DOORBELL_HANDSHAKE		= 1,
	MU_INBOUND_DOORBELL_REQHEADCHANGED	= 2,
	MU_INBOUND_DOORBELL_STATUSTAILCHANGED	= 4,
	MU_INBOUND_DOORBELL_HMUSTOPPED		= 8,
	MU_INBOUND_DOORBELL_RESET		= 16,

	MU_OUTBOUND_DOORBELL_HANDSHAKE		= 1,
	MU_OUTBOUND_DOORBELL_REQUESTTAILCHANGED	= 2,
	MU_OUTBOUND_DOORBELL_STATUSHEADCHANGED	= 4,
	MU_OUTBOUND_DOORBELL_BUSCHANGE		= 8,
	MU_OUTBOUND_DOORBELL_HASEVENT		= 16,

	ST_CAN_QUEUE	= 32,
	ST_MAX_SG	= 32,
	SG_CF_EOT	= 0x80,	/* End of table */
	SG_CF_64B	= 0x40,	/* 64 bit format item */
	SG_CF_HOST	= 0x20,	/* SG is in host address space */
	SG_CF_RAW	= 0x10,	
	SG_CF_DMA	= 0x08,

	ST_MAX_ADAPTORS		= 16,
	ST_INTERNAL_TIMEOUT	= 30,
};

struct st_sgitem {
	u8 ctrl;	/* SG_CF_xxx */
	u8 reserved[3];
	__le32 count;
	__le32 addr;
	__le32 addr_hi;
} __attribute__((packed));

struct st_sgtable {
	__le16 sg_count;
	__le16 max_sg_count; 
	__le32 sz_in_byte;
	struct st_sgitem table[ST_MAX_SG];
} __attribute__((packed));

struct handshake_frame {
	__le32 rb_phy;		/* Request Payload Queue Physical Address */
	__le32 rb_phy_hi;
	__le16 req_sz;		/* Size of each Request Payload */
	__le16 req_cnt;		/* Count of Req the buffer can hold */
	__le16 status_sz;	/* Size of each Status Payload */
	__le16 status_cnt;	/* Count of Status the buffer can hold */
	__le32 hosttime;	/* Seconds from Jan 1, 1970 (GMT) */
	__le32 hosttime_hi;
	u8 partner_type;	/* Who sends this frame */
	u8 reserved0[7];
	__le32 partner_ver_major;		/* Partner Version */
	__le32 partner_ver_minor;
	__le32 partner_ver_oem;
	__le32 partner_ver_build;
	u32 reserved1[4];
} __attribute__((packed));

#define SHASTA_CDB_LENGTH MAX_COMMAND_SIZE

#define REQ_VARIABLE_LEN	1024
#define STATUS_VAR_LEN		128
struct req_msg {
	__le16 tag;
	u8 lun;
	u8 target;
	u8 task_attr;
	u8 task_manage;
	u8 prd_entry;
	u8 payload_sz;	/* payload size in 4-byte */
	u8 cdb[SHASTA_CDB_LENGTH];
	u8 variable[REQ_VARIABLE_LEN];
} __attribute__((packed));

struct status_msg {
	__le16 tag;
	u8 lun;
	u8 target;
	u8 srb_status;
	u8 scsi_status;
	u8 reserved;
	u8 payload_sz;	/* payload size in 4-byte */
	u8 variable[STATUS_VAR_LEN];
} __attribute__((packed));

#define Max(x,y)			(((x) > (y)) ? (x) : (y))
#define Min(x,y)			(((x) < (y)) ? (x) : (y))

#define SRB_STATUS_SUCCESS		1
#define SRB_STATUS_BUSY			0x05
#define SRB_STATUS_INVALID_REQUEST	0x06
#define SRB_STATUS_SELECTION_TIMEOUT	0x0A
#define SRB_SEE_SENSE 			0x84
#define SRB_STATUS_ERROR		0x80

/*
 * SCSI Task Attribute
 */
#define TASK_ATTRIBUTE_SIMPLE		0x0
#define TASK_ATTRIBUTE_ORDERED		0x2
#define TASK_ATTRIBUTE_HEADOFQUEUE	0x1
#define TASK_ATTRIBUTE_ACA		0x4

#define MU_MAX_REQUEST		32
#define TAG_BITMAP_LENGTH	MU_MAX_REQUEST

/* One message wasted
 * Use MU_MAX_REQUEST+1 to handle MU_MAX_REQUEST messages */
#define MU_REQ_COUNT		(MU_MAX_REQUEST + 1)
#define MU_STATUS_COUNT		(MU_MAX_REQUEST + 1)
#define MU_REQ_BUFFER_SIZE	(MU_REQ_COUNT * sizeof(struct req_msg))
#define MU_STATUS_BUFFER_SIZE	(MU_STATUS_COUNT * sizeof(struct status_msg))
#define MU_BUFFER_SIZE		(MU_REQ_BUFFER_SIZE + MU_STATUS_BUFFER_SIZE)

#define PASSTHRU_REQ_HEAD_TYPE 	0x00000001

#define ST_MAX_ARRAY_SUPPORTED	16
#define ST_MAX_TARGET_NUM	(ST_MAX_ARRAY_SUPPORTED+1)
#define ST_MAX_LUN_PER_TARGET	16

#define ST_IOCTL_SIGNATURE1   0x54534146
#define ST_IOCTL_SIGNATURE2   0x4b415254

struct passthru_buf{
	u8 DataBuffer[4096];
};

struct ver_info {
	u32 VerMajor;
	u32 VerMinor;
	u32 VerOEM;
	u32 VerBuild;
	u32	wReserved[2];
};

struct st_frame {
	u32 BaseAddress[6];
	u32 ExpansionROMAddress;

	struct ver_info DriverVersion;
	struct ver_info BIOSVersion;

	u32 SystemIoBusNumber;
	u32 SlotNumber;
	u32 InterruptLevel;
	u32 InterruptVector;
	u32 ID;	
	u32 SubID;

	u32 DIMMSize;
	u8 DIMMType;	
	u8 bReserved[3];
	
	u32 ChannelNumber;
	u32 Reserved;
};

struct st_drvver {
	u32 Major;
	u32 Minor;
	u32 Oem;
	u32 BuildNo;
	u32 Signature[2];
	u8 ConsoleID;
	u8 HostNumber;
	u8 bReserved[2];
	u32 Reserved[3];
};

struct st_ioctl {
	u32 	HeaderLength;
	u32	Signature1;
	u32	Signature2;
	u32	Timeout;
	u32	ControlCode;
	u32	ReturnCode;
	u32	Length;	
};

struct st_passthru {
	u8 SrbStatus;
	u8 ScsiStatus;
	u8 TargetId;
	u8 Lun;
	u8 CdbLength;
	u8 SenseInfoLength;
	u32 DataTransferLength;
	u8 Cdb[16];
	u8 SenseBuffer[32];
};

#define ST_IOCTL_GETVERSION	0x00000001
#define ST_IOCTL_PASSTHRU	_IOWR(0, 1, struct passthru_buf)
#define ST_IOCTL_DATA_SIZE	(0x4000)
#define ST_IOCTL_SIZE (ST_IOCTL_DATA_SIZE + sizeof(struct st_ioctl)	\
			+ sizeof(struct st_passthru))

struct st_ccb {
	struct req_msg *req;
	struct scsi_cmnd *cmd;

	wait_queue_head_t waitq;

	void *sense_buffer;
	void	*data_buffer;
	int data_bufflen;
	int sense_bufflen;
	
	u32 req_type;
	u8 srb_status;
	u8 scsi_status;
};

struct st_hba {
	void __iomem *mmio_base;	/* iomapped PCI memory space */
	void *dma_mem;
	dma_addr_t dma_handle;

	struct Scsi_Host *host;
	struct pci_dev *pdev;
	struct work_struct work_q;

	u32 tag;
	u32 req_head;
	u32 req_tail;
	u32 status_head;
	u32 status_tail;

	struct status_msg *status_buffer;
	struct st_ccb ccb[MU_MAX_REQUEST];
	struct st_ccb *wait_ccb;

	void *ioctl_buffer;
	struct semaphore	ioctl_sem;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
	struct timer_list ioctl_timer;
#endif

	unsigned int mu_status;
	int out_req_cnt;

	u8 irq;
};

static int major = -1;
static int gSt_max_lun = 0;
static int gNumAdapter = 0;
struct st_hba *gShasta_hostp[ST_MAX_ADAPTORS] = {0};

static int shasta_open(struct inode *, struct file *);
static int shasta_ioctl(struct inode *, struct file *, uint, unsigned long);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
static long shasta_compat_ioctl(struct file *, unsigned int, unsigned long);
#endif
static int shasta_handshake(struct st_hba *hba);
static void shasta_hba_stop(struct st_hba *);

static int shasta_init(void);
static void shasta_exit(void);
 
static int shasta_probe(struct pci_dev*, const struct pci_device_id *);
static void shasta_remove(struct pci_dev *);
static void shasta_shutdown(struct device *);
 
static irqreturn_t shasta_intr(int, void *, struct pt_regs *);
 
MODULE_AUTHOR("Ed Lin");
MODULE_DESCRIPTION("Promise Technology SuperTrak EX Controllers");
MODULE_LICENSE("GPL");

static struct file_operations shasta_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= shasta_ioctl,
	.open		= shasta_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	.compat_ioctl	= shasta_compat_ioctl,
#endif
};

static inline void shasta_gettime(u32 *time)
{
	struct timeval tv;
	do_gettimeofday(&tv);

	*time = cpu_to_le32(tv.tv_sec & 0xffffffff);
	*(time + 1) = cpu_to_le32((tv.tv_sec >> 16) >> 16); 
}

static inline u16 shasta_alloc_tag(u32 *bitmap)
{
	u16 i;
	for (i = 0; i < TAG_BITMAP_LENGTH; i++) {
		if (!((*bitmap) & (1 << i))) {
			*bitmap |= (1 << i);
			return i;
		}
	}

	return TAG_BITMAP_LENGTH;
}

static inline void shasta_free_tag(u32 *bitmap, u16 tag)
{
	*bitmap &= ~(1 << tag);
}

static inline struct status_msg *shasta_get_status(struct st_hba *hba)
{
	struct status_msg *pStatus = 
		hba->status_buffer + hba->status_tail;

	if (++hba->status_tail >= MU_STATUS_COUNT)
		hba->status_tail = 0;

	return pStatus;
}

static inline struct req_msg *shasta_alloc_req(struct st_hba *hba)
{
	struct req_msg *req = ((struct req_msg *)hba->dma_mem) +
		hba->req_head;

	if (++hba->req_head >= MU_REQ_COUNT)
		hba->req_head = 0;

	return req;
}

static inline void shasta_map_sg(struct st_hba *hba,
	struct req_msg *req, struct scsi_cmnd *cmd)
{
	struct pci_dev *pdev = hba->pdev;
	dma_addr_t dma_handle;
	struct scatterlist *src;
	struct st_sgtable *dst;
	int i;

	dst = (struct st_sgtable *)req->variable;
	dst->max_sg_count = cpu_to_le16(ST_MAX_SG);

	if (cmd->use_sg) {
		src = (struct scatterlist *) cmd->request_buffer;
		dst->sg_count = cpu_to_le16((u16)pci_map_sg(pdev, src,
			cmd->use_sg, cmd->sc_data_direction));

		for (i = 0; i < dst->sg_count; i++, src++) {
			dst->table[i].count = cpu_to_le32((u32)sg_dma_len(src));
			dst->table[i].addr = 
				cpu_to_le32(sg_dma_address(src) & 0xffffffff);
			dst->table[i].addr_hi = 
				cpu_to_le32((sg_dma_address(src) >> 16) >> 16);
			dst->table[i].ctrl = SG_CF_64B | SG_CF_HOST;
    		}
		dst->table[--i].ctrl |= SG_CF_EOT;			
	}
	else {
#if defined(CONFIG_X86_64)
#if defined(CONFIG_DISCONTIGMEM) 
		dma_handle = pci_map_page(pdev,
			virt_to_page(cmd->request_buffer),
			((unsigned long)cmd->request_buffer & ~PAGE_MASK),
			cmd->request_bufflen, cmd->sc_data_direction);
#else
		dma_handle = pci_map_single(pdev, cmd->request_buffer,
			cmd->request_bufflen, cmd->sc_data_direction);
#endif
#else
	  	dma_handle = pci_map_page(pdev,
	  		virt_to_page(cmd->request_buffer),
	  		((unsigned long)cmd->request_buffer & ~PAGE_MASK),
	  		cmd->request_bufflen,cmd->sc_data_direction);
#endif
		dst->sg_count = cpu_to_le16(1);
		dst->table[0].addr = cpu_to_le32(dma_handle & 0xffffffff);
		dst->table[0].addr_hi = cpu_to_le32((dma_handle >> 16) >> 16);
		dst->table[0].count = cpu_to_le32((u32)cmd->request_bufflen);
		dst->table[0].ctrl = SG_CF_EOT | SG_CF_64B | SG_CF_HOST;

		cmd->SCp.dma_handle = dma_handle;
  	}
	
dst->sz_in_byte = cpu_to_le32(cmd->request_bufflen);
}

static inline void
shasta_send_cmd(struct st_hba *hba, struct req_msg *req, u16 tag)
{
	req->tag = cpu_to_le16(tag);

	req->task_attr = TASK_ATTRIBUTE_SIMPLE;
	req->task_manage = 0; /* not supported yet */
	req->payload_sz = (u8)((sizeof(struct req_msg))/sizeof(u32));

	hba->ccb[tag].req = req;

	hba->out_req_cnt++;

	wmb();
	writel(hba->req_head, hba->mmio_base + IMR0);
	wmb();
	writel(MU_INBOUND_DOORBELL_REQHEADCHANGED, hba->mmio_base + IDBL);
	readl(hba->mmio_base + IDBL); /* flush */
}

static int
shasta_queuecommand(struct scsi_cmnd *cmd, void (* fn)(struct scsi_cmnd *))
{
	struct st_hba *hba;
 	struct Scsi_Host *host;
	unsigned int channel,id,lun;
	struct req_msg *req;
	u16 tag;

	host = cmd->device->host;
	channel = cmd->device->channel;
	id = cmd->device->id;
	lun = cmd->device->lun;

	if (cmd->cmnd[0] == INQUIRY) { /* firmware lun issue work around */
		if (lun > gSt_max_lun)
			gSt_max_lun = lun;
	}

	if (gSt_max_lun == 0) {
		if (channel > 0) {
			lun = channel;
			channel = 0;
		}
	}

	if (channel > 0 ) {
    		cmd->result = (DID_BAD_TARGET << 16);
    		fn(cmd);
    		return 0;
 	 }

	cmd->scsi_done = fn;

	hba = (struct st_hba *)host->hostdata;
	if((tag = shasta_alloc_tag(&hba->tag)) == TAG_BITMAP_LENGTH)
		return SCSI_MLQUEUE_HOST_BUSY;
	req = shasta_alloc_req(hba);

	req->lun = lun;
	req->target = id;

	/* CDB */
	memcpy(req->cdb, cmd->cmnd, SHASTA_CDB_LENGTH);

	if (cmd->request_bufflen != 0)
		shasta_map_sg(hba, req, cmd);

	hba->ccb[tag].cmd = cmd;

	hba->ccb[tag].sense_bufflen = SCSI_SENSE_BUFFERSIZE;
	hba->ccb[tag].sense_buffer = (void *)cmd->sense_buffer;
	hba->ccb[tag].data_bufflen = cmd->request_bufflen;
	hba->ccb[tag].data_buffer = (void *)cmd->request_buffer;

	hba->ccb[tag].req_type = 0;

	shasta_send_cmd(hba, req, tag);

    	return 0;
}

static inline void shasta_unmap_sg(struct st_hba *hba, struct scsi_cmnd *cmd)
{
	dma_addr_t dma_handle;
	if (cmd->sc_data_direction!= PCI_DMA_NONE) {
 		if (cmd->use_sg) {
			pci_unmap_sg(hba->pdev, cmd->request_buffer,
				cmd->use_sg, cmd->sc_data_direction);
		} else {
 			dma_handle = cmd->SCp.dma_handle;
#if defined(CONFIG_X86_64)
#if defined(CONFIG_DISCONTIGMEM)
			pci_unmap_page(hba->pdev, dma_handle, cmd->request_bufflen,
				cmd->sc_data_direction);
#else
			pci_unmap_single(hba->pdev, dma_handle, cmd->request_bufflen,
				cmd->sc_data_direction);
#endif
#else
	  		pci_unmap_page(hba->pdev, dma_handle, cmd->request_bufflen,
	  			cmd->sc_data_direction);
#endif
 		}
	}
}

static inline void shasta_scsi_fin(struct st_ccb *ccb)
{
	struct scsi_cmnd *cmd = ccb->cmd;
	int result;
	switch (ccb->srb_status) {
		case SRB_STATUS_SUCCESS:
			result= DID_OK << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_STATUS_SELECTION_TIMEOUT:
			result= DID_BAD_TARGET << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_SEE_SENSE:
			result= DRIVER_SENSE << 24 | DID_OK << 16 |
				 COMMAND_COMPLETE << 8 | CHECK_CONDITION << 1;
			break;
		case SRB_STATUS_BUSY:
			result = DID_NO_CONNECT << 16 | COMMAND_COMPLETE << 8;
			break;
		default:
			result = DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			break;
  	}

   	cmd->result = result;
   
   	cmd->scsi_done(cmd);
}

static inline void shasta_copy_data(struct st_ccb *ccb,
	struct status_msg *pStatus, unsigned int variable)
{
	if (variable) {
		if (pStatus->scsi_status == SAM_STAT_GOOD) {
			if (ccb->data_buffer != NULL)
				memcpy(ccb->data_buffer, pStatus->variable,
					Min(variable, ccb->data_bufflen));
		} else {
			if (ccb->sense_buffer != NULL)
				memcpy(ccb->sense_buffer, pStatus->variable,
					Min(variable, ccb->sense_bufflen));
		}
	}
}

static inline void shasta_mu_intr(struct st_hba *hba, u32 doorbell)
{
	void __iomem *base = hba->mmio_base;
	struct status_msg *pStatus;
	struct st_ccb *ccb;
	unsigned int size, variable;
	u16 tag;
	
	if (doorbell & MU_OUTBOUND_DOORBELL_STATUSHEADCHANGED) {
		/* Status Payloads */
		hba->status_head = readl(base + OMR1);
		if (hba->status_head >= MU_STATUS_COUNT)
			return;
		if (hba->mu_status == MU_STATE_STARTED && hba->out_req_cnt > 0)
		while (hba->status_tail != hba->status_head) {

			pStatus = shasta_get_status(hba);
			
			tag = le16_to_cpu(pStatus->tag);
			if(tag >= TAG_BITMAP_LENGTH)
				continue;
			if ((hba->tag & (1 << tag)) == 0)
				continue;
			
			size = pStatus->payload_sz * sizeof(u32);
			if (size < sizeof(*pStatus) - STATUS_VAR_LEN ||
				size > sizeof(*pStatus)) 
				continue;

			ccb = &hba->ccb[tag];
			if (ccb->req == NULL)
				continue;

			variable = size + STATUS_VAR_LEN - sizeof(*pStatus);

			shasta_copy_data(ccb, pStatus, variable);
			
			ccb->srb_status = pStatus->srb_status;
			ccb->scsi_status = pStatus->scsi_status;
			ccb->req = NULL;
			if(hba->wait_ccb == ccb)
				hba->wait_ccb = NULL;
			hba->out_req_cnt--;

			if (ccb->req_type & PASSTHRU_REQ_HEAD_TYPE) {
				ccb->req_type = 0;
				if (waitqueue_active(&ccb->waitq))
					wake_up(&ccb->waitq);
				continue;
			}
			
			shasta_free_tag(&hba->tag, tag);
  			shasta_unmap_sg(hba, ccb->cmd);
			shasta_scsi_fin(ccb);
		}
		else hba->status_tail = hba->status_head;
		writel(hba->status_head, base + IMR1);
		readl(base + IMR1);
	}

	if (doorbell & MU_OUTBOUND_DOORBELL_REQUESTTAILCHANGED)
		hba->req_tail = readl(base + OMR0);
}

static irqreturn_t shasta_intr(int irq, void *__hba, struct pt_regs *regs)
{
	struct st_hba *hba = __hba;
	void __iomem *base;
	u32 data;
	unsigned long flags;

	if (!hba || !hba->host) {
		return IRQ_NONE;
	}
	
	spin_lock_irqsave(hba->host->host_lock, flags);
	
	base = hba->mmio_base;

	data = readl(base + ODBL);
	
	if (data == 0) {
		spin_unlock_irqrestore(hba->host->host_lock, flags);
		return (IRQ_NONE);
		}
	
	/* Clear the interrupt. */
	writel(data, base + ODBL);
	readl(base + ODBL);
	
	if (data & MU_OUTBOUND_DOORBELL_HASEVENT) {
		data &= ~MU_OUTBOUND_DOORBELL_HASEVENT;
	}

	/* Bus Change Notification from Firmware */
	if (data & MU_OUTBOUND_DOORBELL_BUSCHANGE ) {
		data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
		schedule_work(&hba->work_q);
	}

	/* MU Interrupt */
	if (data)
		shasta_mu_intr(hba, data);

	spin_unlock_irqrestore(hba->host->host_lock, flags);

	return (IRQ_HANDLED);
}

static int shasta_open(struct inode *inode, struct file *filep)
{
	/*
	 * Only allow superuser to access private ioctl interface
	 */
	if (!capable(CAP_SYS_ADMIN)) return (-EACCES);
	if (MINOR(inode->i_rdev) >= gNumAdapter
		|| MAJOR(inode->i_rdev != major))
		return(-ENODEV);

	return 0;
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
static void shasta_timeout(wait_queue_head_t *waitq_p)
{
   	if (waitqueue_active(waitq_p))
		wake_up(waitq_p);
}

static void shasta_add_timer(struct st_hba *hba, unsigned long data, int timeout)
{
   	hba->ioctl_timer.data = data;
   	hba->ioctl_timer.function = (void (*)(unsigned long))shasta_timeout;
	mod_timer(&hba->ioctl_timer, jiffies + timeout);
}
#endif

static void shasta_controller_info( struct st_hba *hba, struct st_frame *p)
{
	int i;
	p->BaseAddress[0] = pci_resource_start(hba->pdev, 0);
	for ( i=1; i<6; i++ )
		p->BaseAddress[i] = 0;
	p->ExpansionROMAddress = 0;

	p->DriverVersion.VerMajor = VER_MAJOR;
	p->DriverVersion.VerMinor = VER_MINOR;
	p->DriverVersion.VerOEM = OEM;
	p->DriverVersion.VerBuild = BUILD_VER;

	p->SystemIoBusNumber = hba->pdev->bus->number;
	p->SlotNumber = hba->pdev->devfn;
	p->InterruptLevel = 0;
	p->InterruptVector = hba->pdev->irq;
	p->ID = hba->pdev->vendor << 16 | hba->pdev->device;
	p->SubID =
		hba->pdev->subsystem_vendor << 16 | hba->pdev->subsystem_device;

	/* filled by fw, le32 stuff */
	p->BIOSVersion.VerMajor = le32_to_cpu(p->BIOSVersion.VerMajor);
	p->BIOSVersion.VerMinor = le32_to_cpu(p->BIOSVersion.VerMinor);
	p->BIOSVersion.VerOEM = le32_to_cpu(p->BIOSVersion.VerOEM);
	p->BIOSVersion.VerBuild = le32_to_cpu(p->BIOSVersion.VerBuild);

	p->DIMMSize = le32_to_cpu(p->DIMMSize);
	p->ChannelNumber = le32_to_cpu(p->ChannelNumber);
}

static int shasta_ioctl(struct inode *inode, struct file *filep, uint cmd, unsigned long arg)
{
	int ret = 0;
	unsigned int aptno = MINOR(inode->i_rdev);
	struct st_hba *hba = gShasta_hostp[aptno];
	struct st_ioctl *pSrbCtl;	
	struct st_drvver *pVersion;
	struct st_passthru *pSptd;
	struct req_msg *req;
	struct st_sgtable *dst;
	u16 tag;
	void *usr_copy_start, *ker_copy_start;
	unsigned long copy_size;
	unsigned long flags;
	int error;
	u8 cdb1;

	if (down_interruptible(&hba->ioctl_sem)) {
		return (-EINTR);
	}

	error = copy_from_user(hba->ioctl_buffer, (void *)arg,
		sizeof(struct st_ioctl));

	if (error)
		goto err_out;

	pSrbCtl = (struct st_ioctl *)(hba->ioctl_buffer);	

	if (pSrbCtl->Signature1 != ST_IOCTL_SIGNATURE1 ||
		pSrbCtl->Signature2 != ST_IOCTL_SIGNATURE2)
	 	goto err_out;
	
	usr_copy_start = (void *)arg + sizeof(struct st_ioctl);
	ker_copy_start = (void *)(pSrbCtl + 1);

	switch (cmd) {
		case ST_IOCTL_GETVERSION:
			pVersion = (struct st_drvver *)ker_copy_start;
			pVersion->Major = VER_MAJOR;
			pVersion->Minor = VER_MINOR;
			pVersion->Oem = OEM;
			pVersion->BuildNo = BUILD_VER;
			pVersion->Signature[0] = 0x4e415041;
			pVersion->ConsoleID = ST_MAX_ARRAY_SUPPORTED;
			pVersion->HostNumber = hba->host->host_no;
			copy_size = sizeof(struct st_drvver );
			break;
		
		case ST_IOCTL_PASSTHRU:
			pSptd = (struct st_passthru *)ker_copy_start;
			error = copy_from_user(ker_copy_start, usr_copy_start,
				sizeof(struct st_passthru));
			if (error)
				goto err_out;

			if (pSptd->TargetId == ST_MAX_ARRAY_SUPPORTED && pSptd->Lun == 0) {
				ker_copy_start += sizeof(struct st_passthru);
				usr_copy_start += sizeof(struct st_passthru);
				if (pSptd->DataTransferLength > ST_IOCTL_DATA_SIZE) {
					printk("shasta: invalid ioctl buffer size\n");
					goto err_out;
				}
				if(pSptd->DataTransferLength) {
					error = copy_from_user(ker_copy_start, usr_copy_start,
						pSptd->DataTransferLength);
					if (error)
						goto err_out;
				}
				cdb1 = pSptd->Cdb[1];
				spin_lock_irqsave(hba->host->host_lock, flags);
				if((tag = shasta_alloc_tag(&hba->tag)) == TAG_BITMAP_LENGTH){
					spin_unlock_irqrestore(hba->host->host_lock, flags);
					printk("shasta: unable to alloc tag\n");
					goto err_out;
					}
				req = shasta_alloc_req(hba);

				req->lun = pSptd->Lun;
				req->target = pSptd->TargetId;

				/* CDB */
				memcpy(req->cdb, pSptd->Cdb, Min(SHASTA_CDB_LENGTH, pSptd->CdbLength));

				dst = (struct st_sgtable *)(req->variable);

				dst->max_sg_count = cpu_to_le16(ST_MAX_SG);
				dst->sg_count = cpu_to_le16(1);
				dst->table[0].addr = cpu_to_le32(hba->dma_handle
					+ MU_BUFFER_SIZE + sizeof(struct st_ioctl)
					+ sizeof(struct st_passthru));
				dst->table[0].addr_hi = cpu_to_le32((hba->dma_handle >> 16) >> 16);
				dst->table[0].count = cpu_to_le32(pSptd->DataTransferLength);
				dst->table[0].ctrl = SG_CF_EOT | SG_CF_64B | SG_CF_HOST;
  				dst->sz_in_byte = cpu_to_le32(pSptd->DataTransferLength);

				hba->ccb[tag].sense_bufflen = Min(pSptd->SenseInfoLength, 32);
				hba->ccb[tag].sense_buffer = 
					ker_copy_start - Min(pSptd->SenseInfoLength,32);
				hba->ccb[tag].data_bufflen = pSptd->DataTransferLength;
				hba->ccb[tag].data_buffer = ker_copy_start;

				hba->ccb[tag].req_type = PASSTHRU_REQ_HEAD_TYPE;

				shasta_send_cmd(hba, req, tag);

				init_waitqueue_head(&(hba->ccb[tag].waitq));	
				copy_size = sizeof(struct st_passthru)
					+ pSptd->DataTransferLength;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
				shasta_add_timer(hba, (unsigned long)&hba->ccb[tag].waitq,
					ST_INTERNAL_TIMEOUT*HZ);
#endif
				spin_unlock_irqrestore(hba->host->host_lock, flags);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
				wait_event(hba->ccb[tag].waitq,!(hba->ccb[tag].req_type));
				del_timer_sync(&hba->ioctl_timer);
#else
				wait_event_timeout(hba->ccb[tag].waitq,
					!(hba->ccb[tag].req_type), ST_INTERNAL_TIMEOUT*HZ);
#endif

				if (hba->ccb[tag].req_type == PASSTHRU_REQ_HEAD_TYPE) {
					pSptd->SrbStatus = SRB_STATUS_SELECTION_TIMEOUT;
				} else {
					pSptd->SrbStatus = hba->ccb[tag].srb_status;
					pSptd->ScsiStatus = hba->ccb[tag].scsi_status;
					spin_lock_irqsave(hba->host->host_lock, flags);
					shasta_free_tag(&hba->tag, tag);
					spin_unlock_irqrestore(hba->host->host_lock, flags);
				}
				if ( cdb1 == 5 ) {
					shasta_controller_info(hba, (struct st_frame*)(ker_copy_start));
				}
				ker_copy_start -= sizeof(struct st_passthru);
				usr_copy_start -= sizeof(struct st_passthru);

				if( (SRB_STATUS_SUCCESS == pSptd->SrbStatus)
					|| (SRB_SEE_SENSE == pSptd->SrbStatus) )
					ret = 0;
				else
					ret = -EFAULT;

			} /* if console request. */
			else goto err_out;
			break;
		
		default:
			ret = -EINVAL;
			goto out;
	}	

	error = copy_to_user(usr_copy_start, ker_copy_start, copy_size);

	if (error )
		ret = -EFAULT;
out:
	up(&hba->ioctl_sem);
	return ret;	
err_out:
	ret = -EFAULT;
	goto out;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
/* CONFIG_COMPAT */
static long
shasta_compat_ioctl(struct file *filep, unsigned int cmd, unsigned long arg)
{
	int err;

	err = shasta_ioctl(filep->f_dentry->d_inode, filep, cmd, arg);

	return err;
}
#endif

static void shasta_bus_change(void *__hba)
{
}

static int shasta_abort(struct scsi_cmnd *cmd)
{
	struct st_hba *hba;
	struct Scsi_Host *host;
	u16 tag;
	void __iomem *base;
	u32 Data;
	unsigned long flags;
	if (!cmd) {
		printk(KERN_WARNING "shasta_abort(): Invalid Scsi_Cmnd\n");
		return (FAILED);
	}

	host = cmd->device->host;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	spin_lock_irqsave(host->host_lock, flags);
#endif
	hba = (struct st_hba *)host->hostdata;

	base = hba->mmio_base;

	for(tag = 0; tag < MU_MAX_REQUEST; tag++)
		if (hba->ccb[tag].cmd == cmd) {
			hba->wait_ccb = &(hba->ccb[tag]);
			break;
		}
	if(tag >= MU_MAX_REQUEST){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
		spin_unlock_irqrestore(host->host_lock, flags);
#endif
		return FAILED;
	}

	Data = readl(base + ODBL);
	if(Data == 0) {
		hba->wait_ccb = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
		spin_unlock_irqrestore(host->host_lock, flags);
#endif
		return FAILED;
	}

	writel(Data, base + ODBL);
	readl(base + ODBL);

	if ( Data&MU_OUTBOUND_DOORBELL_HASEVENT ) {
		Data &=~MU_OUTBOUND_DOORBELL_HASEVENT;
		}

	if ( Data&MU_OUTBOUND_DOORBELL_BUSCHANGE ) {
		Data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
		schedule_work(&hba->work_q);
		}

	if ( Data!=0 )
		shasta_mu_intr(hba, Data);

	if(hba->wait_ccb == NULL){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
		spin_unlock_irqrestore(host->host_lock, flags);
#endif
		return SUCCESS;
		}

	hba->wait_ccb = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	spin_unlock_irqrestore(host->host_lock, flags);
#endif
	return FAILED;
}

static int shasta_reset(struct scsi_cmnd *cmd)
{
	struct st_hba *hba;
	struct Scsi_Host *host;
	unsigned long flags;
	if (!cmd) {
		printk(KERN_WARNING "shasta_reset(): Invalid Scsi_Cmnd\n");
		return (FAILED);
	}

	host = cmd->device->host;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	spin_lock_irqsave(host->host_lock, flags);
#endif
	hba = (struct st_hba *)host->hostdata;

	if (shasta_handshake(hba)) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
		spin_unlock_irqrestore(host->host_lock, flags);
#endif
		printk(KERN_WARNING "shasta_reset(): handshake failed\n");
		return SUCCESS;
		}
	hba->tag = 0;
	hba->req_head = 0;
	hba->req_tail = 0;
	hba->status_head = 0;
	hba->status_tail = 0;
	hba->out_req_cnt = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	spin_unlock_irqrestore(host->host_lock, flags);
#endif
	return SUCCESS;
}

static int shasta_handshake(struct st_hba *hba)
{
	void __iomem *base = hba->mmio_base;
	struct handshake_frame *h;
	int i;

	if (readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE) {
		writel(MU_INBOUND_DOORBELL_HANDSHAKE, base + IDBL);
		readl(base + IDBL);
		for (i = 0; readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE
			&& i < MU_MAX_DELAY_TIME; i++) {
			rmb();
			msleep(1);
		}

		if (i == MU_MAX_DELAY_TIME) {
			printk(KERN_ERR SHASTA_NAME ": no handshake signature\n");
			return -1;
		}
	}

	udelay(10);

	h = (struct handshake_frame *)(hba->dma_mem + MU_REQ_BUFFER_SIZE);

	h->rb_phy = cpu_to_le32(hba->dma_handle);
	h->rb_phy_hi = cpu_to_le32((hba->dma_handle >> 16) >> 16);
	h->req_sz = cpu_to_le16(sizeof(struct req_msg));
	h->req_cnt = cpu_to_le16(MU_REQ_COUNT);
	h->status_sz = cpu_to_le16(sizeof(struct status_msg));
	h->status_cnt = cpu_to_le16(MU_STATUS_COUNT);
	shasta_gettime(&h->hosttime);
	h->partner_type = HMU_PARTNER_TYPE;

	wmb();
	writel(hba->dma_handle + MU_REQ_BUFFER_SIZE, base + IMR0);
	wmb();
	writel((hba->dma_handle >> 16) >> 16, base + OMR0); /* 4G border safe */
	wmb();
	writel(MU_INBOUND_DOORBELL_HANDSHAKE, base + IDBL);
	readl(base + IDBL);

	udelay(10);
	for (i = 0; readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE
		&& i < MU_MAX_DELAY_TIME; i++) {
		rmb();
		msleep(1);
	}

	if (i == MU_MAX_DELAY_TIME) {
		printk(KERN_ERR SHASTA_NAME
			": no signature after handshake frame\n");
		return -1;
	}

	writel(0, base + IMR0);
	readl(base + IMR0);
	writel(0, base + OMR0);
	readl(base + OMR0);
	writel(0, base + IMR1);
	readl(base + IMR1);
	writel(0, base + OMR1);
	readl(base + OMR1);
	hba->mu_status = MU_STATE_STARTED;
	return 0;
}

static void shasta_init_hba(struct st_hba *hba,
	struct Scsi_Host * host, struct pci_dev *pdev)
{
	host->max_channel = ST_MAX_LUN_PER_TARGET;
    	host->max_id = ST_MAX_TARGET_NUM;
    	host->max_lun = ST_MAX_LUN_PER_TARGET;
	host->unique_id = host->host_no;
	host->max_cmd_len = SHASTA_CDB_LENGTH;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4) && \
	LINUX_VERSION_CODE  < KERNEL_VERSION(2,5,0)
	scsi_set_pci_device(host, pdev);
#endif

	hba->host = host;
	hba->pdev = pdev;
	hba->irq = pdev->irq;
#if LINUX_VERSION_CODE  <= KERNEL_VERSION(2,6,8)
	init_timer(&hba->ioctl_timer);
#endif
	init_MUTEX(&hba->ioctl_sem);
	INIT_WORK(&hba->work_q, shasta_bus_change, (void *)hba);

}

static int shasta_init_shm(struct st_hba *hba, struct pci_dev *pdev)
{
	hba->dma_mem = pci_alloc_consistent(pdev,
		MU_BUFFER_SIZE + ST_IOCTL_SIZE, &hba->dma_handle);
	if (!hba->dma_mem)
		return -ENOMEM;

	hba->status_buffer =
		(struct status_msg *)(hba->dma_mem + MU_REQ_BUFFER_SIZE);
	hba->ioctl_buffer = (hba->dma_mem + MU_BUFFER_SIZE);
	hba->mu_status = MU_STATE_STARTING;
	return 0;
}

static int shasta_biosparam(struct scsi_device *sdev,
	struct block_device *bdev, sector_t capacity, int geom[])
{
	int heads = 255, sectors = 63, cylinders;

	if (capacity < 0x200000) {
		heads = 64;
		sectors = 32;
	}

	cylinders = sector_div(capacity, heads * sectors);

	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;

	return 0;
}

#ifdef _SUPPORT_DISKDUMP

static int shasta_sanity_check(struct scsi_device *sdev)
{
	struct st_hba *hba;

	if (!sdev->host || !sdev->host->hostdata) {
		return -ENXIO;
	}

	hba = (struct st_hba *)sdev->host->hostdata;

	if (hba->mu_status != MU_STATE_STARTED) {
		return -EIO;
	}

	return 0;
}

static int shasta_dump_shutdown(struct scsi_device *sdev)
{
	struct st_hba *hba;

	if (!sdev->host || !sdev->host->hostdata) {
		return -ENXIO;
	}

	hba = (struct st_hba *)sdev->host->hostdata;

	shasta_hba_stop(hba);

	return 0;
}

static void shasta_poll(struct scsi_device *sdev)
{
	struct st_hba *hba;
	void __iomem *base;
	u32 data;

	if (!sdev->host || !sdev->host->hostdata) {
		return;
	}

	hba = (struct st_hba *)sdev->host->hostdata;

	spin_lock(hba->host->host_lock);

	base = hba->mmio_base;

	data = readl(base + ODBL);

	if (data == 0) {
		spin_unlock(hba->host->host_lock);
		return;
		}

	writel(data, base + ODBL);
	readl(base + ODBL);

	if (data & MU_OUTBOUND_DOORBELL_HASEVENT) {
		data &= ~MU_OUTBOUND_DOORBELL_HASEVENT;
	}

	if (data & MU_OUTBOUND_DOORBELL_BUSCHANGE ) {
		data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
	}

	/* MU Interrupt */
	if (data)
		shasta_mu_intr(hba, data);

	spin_unlock(hba->host->host_lock);

	return;
}
#endif

static struct scsi_host_template driver_template = {
	.module				= THIS_MODULE,
	.name				= "Promise SuperTrak EX driver",
	.proc_name			= SHASTA_NAME,
	.bios_param			= shasta_biosparam,
	.queuecommand			= shasta_queuecommand,
	.eh_abort_handler		= shasta_abort,
	.eh_device_reset_handler	= shasta_reset,
	.eh_bus_reset_handler		= shasta_reset,
	.eh_host_reset_handler		= shasta_reset,
	.can_queue			= ST_CAN_QUEUE - 1,
	.this_id			= -1,
	.sg_tablesize			= ST_MAX_SG,
	.cmd_per_lun			= ST_CAN_QUEUE - 1,
#ifdef _SUPPORT_DISKDUMP
	.dump_sanity_check		= shasta_sanity_check,
	.dump_shutdown			= shasta_dump_shutdown,
	.dump_poll			= shasta_poll,
#endif
};

static int shasta_set_dma_mask(struct pci_dev * pdev)
{
#if defined(_64BPLATFORM) || defined(CONFIG_HIGHMEM64G)
		if (pci_set_dma_mask(pdev, 0xffffffffffffffffULL) == 0) {
			if (pci_set_consistent_dma_mask(pdev,
			    0xffffffffffffffffULL)) {
			    	printk(KERN_ERR SHASTA_NAME ": 64 bit PCI consistent mask failed\n");
				return(pci_set_consistent_dma_mask(pdev,
				    0xffffffffULL) );
			}
			return 0;
		} else {
			printk(KERN_ERR SHASTA_NAME ": using 32 bit PCI consistent mask\n");
			return(pci_set_dma_mask(pdev, 0xffffffffULL) );
		}
#else
	return(pci_set_dma_mask(pdev, 0xffffffffULL) );
#endif
}

static int __devinit
shasta_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct st_hba *hba;
	struct Scsi_Host *host;
	int err;

	err = pci_enable_device(pdev);
	if (err)
		return err;

	pci_set_master(pdev);

	host = scsi_host_alloc(&driver_template, sizeof(struct st_hba));

	if (!host) {
		printk(KERN_ERR SHASTA_NAME "(%s): scsi_host_alloc failed\n",
			pci_name(pdev));
		err = -ENOMEM;
		goto out_disable;
	}

	hba = (struct st_hba *)host->hostdata;
	memset(hba, 0, sizeof(struct st_hba));

	err = pci_request_regions(pdev, SHASTA_NAME);
	if (err < 0) {
		printk(KERN_ERR SHASTA_NAME "(%s): request regions failed\n",
			pci_name(pdev));
		goto out_scsi_host_put;
	}

	hba->mmio_base = ioremap(pci_resource_start(pdev, 0),
				pci_resource_len(pdev, 0));
	if ( !hba->mmio_base) {
		printk(KERN_ERR SHASTA_NAME "(%s): memory map failed\n",
			pci_name(pdev));
		err = -ENOMEM;
		goto out_release_regions;
	}

	err = shasta_set_dma_mask(pdev);
	if (err) {
		printk(KERN_ERR SHASTA_NAME "(%s): set dma mask failed\n",
			pci_name(pdev));
		goto out_iounmap;
	}

	err = shasta_init_shm(hba, pdev);
	if (err) {
		printk(KERN_ERR SHASTA_NAME "(%s): dma mem alloc failed\n",
		       pci_name(pdev));
		goto out_iounmap;
	}

	shasta_init_hba(hba, host, pdev);

	err = request_irq(pdev->irq, shasta_intr, SA_SHIRQ, SHASTA_NAME, hba);
	if (err) {
		printk(KERN_ERR SHASTA_NAME "(%s): request irq failed\n",
		       pci_name(pdev));
		goto out_pci_free;
	}

	err = shasta_handshake(hba);
	if (err)
		goto out_free_irq;

	pci_set_drvdata(pdev, hba);

	err = scsi_add_host(host, &pdev->dev);
	if (err) {
		printk(KERN_ERR SHASTA_NAME "(%s): scsi_add_host failed\n",
			pci_name(pdev));
		goto out_free_irq;
	}

	gShasta_hostp[gNumAdapter++] = hba;

	scsi_scan_host(host);

	if (major == -1) {
		major = register_chrdev(0, SHASTA_NAME, &shasta_fops);
		if (major < 0) {
			printk(KERN_WARNING SHASTA_NAME
				": failed to register char device\n");
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#ifdef __x86_64__
		else
			register_ioctl32_conversion(ST_IOCTL_PASSTHRU, NULL);
#endif
#endif
	}
	return 0;

out_free_irq:
	free_irq(pdev->irq, hba);
out_pci_free:
	pci_free_consistent(pdev, MU_BUFFER_SIZE,
		hba->dma_mem, hba->dma_handle);
out_iounmap:
	iounmap(hba->mmio_base);	
out_release_regions:
	pci_release_regions(pdev);
out_scsi_host_put:
	scsi_host_put(host);
out_disable:
	pci_disable_device(pdev);

	return err;
}

static void shasta_hba_stop(struct st_hba *hba)
{
	struct req_msg *req;
	unsigned long flags;
	int i;
	u16 tag;

	spin_lock_irqsave(hba->host->host_lock, flags);
	for (i=0; i<ST_MAX_ARRAY_SUPPORTED*2; i++) {
		if((tag = shasta_alloc_tag(&hba->tag)) == TAG_BITMAP_LENGTH){
			spin_unlock_irqrestore(hba->host->host_lock, flags);
			printk("shasta: unable to alloc tag\n");
			return;
			}
		req = shasta_alloc_req(hba);

		memset(req->cdb, 0, MAX_COMMAND_SIZE);

		if (i < ST_MAX_ARRAY_SUPPORTED) {
			req->target = i;
			req->lun = 0;
			req->cdb[0] = 0xe1; /*svucCONTROLLER_COMMAND;*/
			req->cdb[1] = 0x0e; /*ccmdPOWER_STATE_CHANGING;*/
			req->cdb[2] = 0x01; /*CCMD_POWER_SAVING;*/
		} else {
			req->target = i -ST_MAX_ARRAY_SUPPORTED;
			req->lun = 0;
   			req->cdb[0] = SYNCHRONIZE_CACHE;
		}

		hba->ccb[tag].cmd = NULL;

		hba->ccb[tag].data_bufflen = 0;
 		hba->ccb[tag].data_buffer = NULL;

 		hba->ccb[tag].sense_bufflen = 0;
		hba->ccb[tag].sense_buffer = NULL;

		hba->ccb[tag].req_type = PASSTHRU_REQ_HEAD_TYPE;

		init_waitqueue_head(&hba->ccb[tag].waitq);

		shasta_send_cmd(hba, req, tag);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
		shasta_add_timer(hba, (unsigned long)&hba->ccb[tag].waitq,
			ST_INTERNAL_TIMEOUT*HZ);
#endif
		spin_unlock_irqrestore(hba->host->host_lock, flags);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
		wait_event(hba->ccb[tag].waitq,!(hba->ccb[tag].req_type));
		del_timer_sync(&hba->ioctl_timer);
#else
		wait_event_timeout(hba->ccb[tag].waitq,
			!(hba->ccb[tag].req_type), ST_INTERNAL_TIMEOUT*HZ);
#endif
		if (hba->ccb[tag].req_type == PASSTHRU_REQ_HEAD_TYPE)
			return;
		spin_lock_irqsave(hba->host->host_lock, flags);
		shasta_free_tag(&hba->tag, tag);
	}

	spin_unlock_irqrestore(hba->host->host_lock, flags);
}

static void shasta_hba_free(struct st_hba *hba)
{
	flush_scheduled_work();

	free_irq(hba->irq, hba);

	iounmap(hba->mmio_base);

	pci_release_regions(hba->pdev);

	if(hba->dma_mem)
		pci_free_consistent(hba->pdev, MU_BUFFER_SIZE,
			hba->dma_mem, hba->dma_handle);
}

static void shasta_remove(struct pci_dev *pdev)
{
	struct st_hba *hba;

	hba = pci_get_drvdata(pdev);

	if (!hba) {
		printk(KERN_CRIT SHASTA_NAME "(%s): invalid detach\n",
		       pci_name(pdev));

		return;
	}

	scsi_remove_host(hba->host);

	if (major >= 0) {
		unregister_chrdev(major, SHASTA_NAME);
		major = -1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#ifdef __x86_64__
		unregister_ioctl32_conversion(ST_IOCTL_PASSTHRU);
#endif
#endif
	}

	pci_set_drvdata(pdev, NULL);

	shasta_hba_stop(hba);

	shasta_hba_free(hba);

	scsi_host_put(hba->host);

	pci_disable_device(pdev);
}

static void shasta_shutdown(struct device *device)
{
	struct st_hba *hba = pci_get_drvdata(to_pci_dev(device));

	if (!hba) {
		printk(KERN_CRIT SHASTA_NAME "(%s): invalid shutdown\n",
		       pci_name(to_pci_dev(device)));
		return;
	}

	shasta_hba_stop(hba);
}

static struct pci_device_id shasta_pci_tbl[] = {
	{ PCI_VENDOR_ID_INTEL, 0x0334, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
	{ PCI_VENDOR_ID_INTEL, 0x0374, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
	{ PCI_VENDOR_ID_PROMISE, 0x8350, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
	{ PCI_VENDOR_ID_PROMISE, 0xf350, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
	{ }	/* terminate list */
};
MODULE_DEVICE_TABLE(pci, shasta_pci_tbl);

static struct pci_driver shasta_pci_driver = {
	.name		= SHASTA_NAME,
	.id_table	= shasta_pci_tbl,
	.probe		= shasta_probe,
	.remove		= shasta_remove,
	.driver		= {
		.shutdown	= shasta_shutdown,
	}
};

static int __init shasta_init(void)
{
	printk(KERN_INFO SHASTA_NAME
		": Promise SuperTrak EX Driver version: %s %s\n",
		 ST_DRIVER_VERSION, ST_DRIVER_DATE);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
	return pci_module_init(&shasta_pci_driver);
#else
	if (!(pci_register_driver(&shasta_pci_driver)) ) {
        	pci_unregister_driver(&shasta_pci_driver);
        	return -ENODEV;
    	}

	return 0;
#endif
}

static void __exit shasta_exit(void)
{
	pci_unregister_driver(&shasta_pci_driver);
}

module_init(shasta_init);
module_exit(shasta_exit);
